import pytest
import numpy as np 
from .. import parameters 

def test_wgr_get_parameters():
    assert type(parameters.wgr_get_parameters(np.random.random(np.random.randint(1, 100)), np.random.uniform(0, 5))) == type(np.asarray([]))
    assert parameters.wgr_get_parameters(np.random.random(np.random.randint(1, 100)), np.random.uniform(0, 5)).size == 3
    assert np.allclose(parameters.wgr_get_parameters(np.zeros(np.random.randint(1, 100)), np.random.uniform(0, 5)), np.zeros(3))
    dt = np.random.uniform(0, 5)
    assert np.allclose(parameters.wgr_get_parameters(np.ones(np.random.randint(100)), dt), np.asarray([1., dt, dt]))
    assert np.allclose(parameters.wgr_get_parameters(np.asarray([0.54353434, 0.39763409, 0.92579636, 0.74797229, 0.89733085]), 0.1), np.asarray([0.92579636, 0.3, 0.1]))


import os
import time
from scipy.io   import savemat
from tkinter    import Tk, Button, Label, OptionMenu, StringVar

from .inputWindow          import InputWindow
from .loggingWindow        import LoggingWindow
from .plotterOptionsWindow import PlotterOptionsWindow
from .parameterWindow      import ParameterWindow
     
from ..misc.status import Status
from ..core.core   import Core 

class Main():
    def __init__(self):
        # main window
        root = Tk()
        root.title("rsHRF Toolbox")
        # get screen width and height
        screen_width     = root.winfo_screenwidth()
        screen_height    = root.winfo_screenheight()
        # placing the toplevel
        root.geometry("500x300+%d+%d" % (((280/1900)*screen_width + 425), (((1040.0-220)/1000)*screen_height)-670))
        # defining the toplevels
        input_window     = InputWindow()
        parameter_window = ParameterWindow()
        core             = Core()
        logger           = LoggingWindow()
        plotter          = PlotterOptionsWindow()
        # defining variables used by tkinter widgets
        current          = StringVar()
        current_subject  = StringVar()
        output_path      = os.getcwd()
        self.options     = ["None"]
        self.subjects    = ["None"]
        current.set        ("None")
        current_subject.set("None")
        # other variables
        input            = ()    # receives the input from the input window
        output           = {}    # receives the output from the core 
        # initializing parameter window
        parameter_window.setParameters(core.get_parameters())
        parameter_window.display()

        ''' Gets the input from the input toplevel.
            The input comprises of three elements:
                1. Input file (resting-state fMRI)
                2. Brain Mask
                3. Estimation Rule
        '''
        def makeInput():
            start                    = time.process_time()
            # interacting with the input window
            input                    = input_window.getInput()
            # interacting with the back-end
            parameters               = parameter_window.getParameters()
            parameters['estimation'] = input[-2]
            core.updateParameters(parameters)      # only passing the estimation-rule
            output                   = core.makeInput(input[:-2])  # receiving input from core
            output_path              = input[-1]                      # the path to output directory
            # interacting with the parameter-window
            parameter_window.setParameters(core.get_parameters())
            parameter_window.display()
            # updating data
            updateSubjects()
            # logging
            time_taken                  = time.process_time() - start
            output.set_time(str(time_taken)[:5])
            logger.putLog(status=output)

        def changeEstimationRule():
            start = time.process_time()
            # interacting with the input window
            input                       = input_window.getInput()
            # interacting with the core
            output                      = core.updateParameters({"estimation":input[-2]})
            # interacting with the parameter-window
            parameter_window.setParameters(core.get_parameters())
            parameter_window.display()
            # logging
            time_taken                  = time.process_time() - start
            output.set_time(str(time_taken)[:5])
            logger.putLog(status=output)

        def updateParameters():
            start                       = time.process_time()
            # interacting with the core
            output                      = core.updateParameters(dic=parameter_window.updateParameters())
            # interacting with the parameter-window
            parameter_window.setParameters(core.get_parameters())
            parameter_window.display()
            time_taken                  = time.process_time() - start
            # logging
            output.set_time(str(time_taken)[:5])
            logger.putLog(status=output)
        
        def retrieveHRF(get_pos=False):
            if "Preprocessed-BOLD" not in current.get():
                logger.putLog(status=Status(False, error="Select a Preprocessed-BOLD Timeseries to retrieve HRF"))
            else:
                start                       = time.process_time()
                core.updateParameters(dic   =parameter_window.updateParameters())
                bold_ts                     = core.get_time_series(current.get())
                output                      = core.retrieveHRF(bold_ts, get_pos)
                # updating data
                updateValues(log=False)
                # logging
                time_taken                  = time.process_time() - start
                if get_pos:
                    output, pos = output 
                else:
                    pos         = None
                output.set_time(str(time_taken)[:5])
                logger.putLog(status=output)
                return pos

        def retrieveHRFParameters(pos=None):
            if pos != None:
                current.set(current.get().split("_")[0] + "_HRF_" + str(pos))
            if "HRF" not in current.get():
                logger.putLog(status=Status(False, error="Select an HRF Timeseries to retrieve parameters"))
            else:
                start                       = time.process_time()
                # interacting with the core
                output                      = core.getHRFParameters(core.get_time_series(current.get()))
                # updating data
                updateValues(log=False)
                # logging
                time_taken                  = time.process_time() - start 
                output.set_time(str(time_taken)[:5])
                logger.putLog(status=output)

        def deconvolveHRF(pos=None):
            if pos != None:
                current.set(current.get().split("_")[0] + "_HRF_" + str(pos))
            if "HRF" not in current.get():
                logger.putLog(status=Status(False, error="Select an HRF Timeseries to deconvolve BOLD"))
            else:
                start                       = time.process_time()
                # interacting with the core
                output                      = core.deconvolveHRF(core.get_time_series(current.get()))
                # updating data
                updateValues(log=False)
                # logging
                time_taken                  = time.process_time() - start 
                output.set_time(str(time_taken)[:5])
                logger.putLog(status=output)

        def runCompletePipeline():
            if "Preprocessed-BOLD" in current.get():
                pos = retrieveHRF(get_pos=True)
                retrieveHRFParameters(pos)
                deconvolveHRF(pos)
            elif "HRF" in current.get():
                retrieveHRFParameters()
                deconvolveHRF()
            else:
                logger.putLog(status=Status(False, error="Select an HRF or Preprocessed-BOLD Timeseries to run pipeline"))

        def updatePlots():
            # interacting with the core
            try:
                plotables                   = core.get_plotables(current_subject.get())
            except: 
                logger.putLog(status=Status(False, error="Please select a subject"))
            else:
                # logging
                if plotables == None:
                    logger.putLog(status=Status(False, error="Please select a subject"))
                else:
                    logger.putLog(plotables=plotables)
                    # interacting with the plotter
                    plotter.updatePlots(plotables)

        def getInfo():
            try:
                if len(current.get().split("_")) == 3:
                    logger.putLog(data_info=core.get_store_info(current.get()))
                else:
                    logger.putLog(status = Status(False, error="Select a valid input to get information")) 
            except: 
                logger.putLog(status = Status(False, error="Select a valid input to get information")) 
   
        def saveValue():
            try:
                if len(current.get().split("_")) == 3:
                    if os.path.isdir(output_path):
                        out = output_path 
                    else:
                        out = os.cwd() 
                    if core.save_info(current.get(), out):
                        logger.putLog(status = Status(True, info="File saved successfully"))   
                    else:
                        logger.putLog(status = Status(False, error="Unable to save file"))      
                else:
                    logger.putLog(status = Status(False, error="Select a valid input to save")) 
            except:
                logger.putLog(status = Status(False, error="Select a valid input to save")) 

        def saveForSubject():
            try:
                temp                        = core.get_data_labels(current_subject.get())
            except:
                logger.putLog(status=Status(False, error="Please select a subject")) 
            else:
                for each in temp:
                    current.set(each)
                    saveValue()
        
        def add_new(l1, l2):
            if l1 == None:
                return l2 
            if l2 == None:
                return l1
            out = []
            for each in l1:
                if each not in l2:
                    out.append(each)
            for each in l2:
                out.append(each)
            return out
        
        def updateSubjects():
            temp                        = list(core.get_subjects())
            temp.append                 ("None")
            self.subjects               = add_new(self.subjects, temp)
            self.subjects               = sorted(self.subjects)
            subjectOptions              = OptionMenu(root, current_subject, *self.subjects)
            subjectOptions.grid         (row=0,column=1,padx=(30,30),pady=(5,5))
        
        def updateValues(log=True):
            try:
                temp                        = core.get_data_labels(current_subject.get())
            except:
                logger.putLog(status=Status(False, error="Please select a subject"))
            else:
                if log:
                    logger.putLog(status=Status(True, info="Updated values for subject: " + current_subject.get()))
                temp.append             ("None")
                self.options            = add_new(self.options, temp)
                self.options            = sorted(self.options)
                valueOptions            = OptionMenu(root, current,         *self.options)
                valueOptions.grid       (row=1,column=1,padx=(30,30),pady=(5,5))

       

        # defining the widgets
        getInput                        = Button    (root, text="Get Input",               command=makeInput,             height = 1, width = 20)
        changeParameterButton           = Button    (root, text="Update Parameters",       command=updateParameters,      height = 1, width = 20)
        retrieveHRFButton               = Button    (root, text="Retrieve HRF",            command=retrieveHRF,           height = 1, width = 20)
        retrieveHRFParametersButton     = Button    (root, text="Retrieve HRF Parameters", command=retrieveHRFParameters, height = 1, width = 20)
        deconvolveHRFButton             = Button    (root, text="Deconvolve BOLD",         command=deconvolveHRF,         height = 1, width = 20)
        updatePlotsButton               = Button    (root, text="Update Plots",            command=updatePlots,           height = 1, width = 20)
        changeEstimationRule            = Button    (root, text="Set Estimation Rule",     command=changeEstimationRule,  height = 1, width = 20)
        getValueInfo                    = Button    (root, text="Get Info",                command=getInfo,               height = 1, width = 20)
        storeValue                      = Button    (root, text="Save Value",              command=saveValue,             height = 1, width = 20)
        updateStore                     = Button    (root, text="Update Values",           command=updateValues,          height = 1, width = 20)
        runCompletePipeline             = Button    (root, text="Run Pipeline",            command=runCompletePipeline,   height = 1, width = 20)
        saveAllValues                   = Button    (root, text="Save All",                command=saveForSubject,        height = 1, width = 20)
        dataLabel                       = Label     (root, text="Stored Values: ")
        subjectLabel                    = Label     (root, text="Subjects: ")
        valueOptions                    = OptionMenu(root, current,         *self.options)
        subjectOptions                  = OptionMenu(root, current_subject, *self.subjects)
        # placing the widgets
        subjectLabel.grid               (row=0,column=0,padx=(30,30),pady=(5,5))
        subjectOptions.grid             (row=0,column=1,padx=(30,30),pady=(5,5))
        dataLabel.grid                  (row=1,column=0,padx=(30,30),pady=(5,5))
        valueOptions.grid               (row=1,column=1,padx=(30,30),pady=(5,5))
        getInput.grid                   (row=2,column=0,padx=(30,30),pady=(5,5))
        updateStore.grid                (row=2,column=1,padx=(30,30),pady=(5,5))
        changeEstimationRule.grid       (row=3,column=0,padx=(30,30),pady=(5,5))
        changeParameterButton.grid      (row=3,column=1,padx=(30,30),pady=(5,5))
        retrieveHRFButton.grid          (row=4,column=0,padx=(30,30),pady=(5,5))
        retrieveHRFParametersButton.grid(row=4,column=1,padx=(30,30),pady=(5,5))
        deconvolveHRFButton.grid        (row=5,column=0,padx=(30,30),pady=(5,5))
        updatePlotsButton.grid          (row=5,column=1,padx=(30,30),pady=(5,5))
        getValueInfo.grid               (row=6,column=0,padx=(30,30),pady=(5,5))
        storeValue.grid                 (row=6,column=1,padx=(30,30),pady=(5,5))
        runCompletePipeline.grid        (row=7,column=0,padx=(30,30),pady=(5,5))
        saveAllValues.grid              (row=7,column=1,padx=(30,30),pady=(5,5))

        root.mainloop()
    

if __name__ == "__main__":
    Main()


import math
import numpy   as np
import nibabel as nib
from scipy.special import gammaln

import warnings
warnings.filterwarnings("ignore")

def spm_vol(input_file):
    """
    Get header information for images
    """
    v = nib.load(input_file)
    return v


def spm_read_vols(mapped_image_volume):
    """
    Read in entire image volumes
    """
    data = mapped_image_volume.get_data()
    data = data.flatten(order='F')
    return data


def spm_orth(X, OPT='pad'):
    """
    Recursive Gram-Schmidt orthogonalisation of basis functions
    @X - matrix
    @OPT - 'norm' - for Euclidean normalisation
           'pad'  - for zero padding of null space (default)
    """
    def gs_cofficient(v1, v2):
        return np.dot(v2, v1) / np.dot(v1, v1)

    def multiply(cofficient, v):
        return map((lambda x: x * cofficient), v)

    def proj(v1, v2):
        return multiply(gs_cofficient(v1, v2), v1)

    def gs(X, row_vecs=True, norm=True):
        if not row_vecs:
            X = X.T
        Y = X[0:1, :].copy()
        for i in range(1, X.shape[0]):
            proj = np.diag((X[i, :].dot(Y.T) /
                            np.linalg.norm(Y, axis=1) ** 2).flat).dot(Y)
            Y = np.vstack((Y, X[i, :] - proj.sum(0)))
        if norm:
            Y = np.diag(1 / np.linalg.norm(Y, axis=1)).dot(Y)
        if row_vecs:
            return Y
        else:
            return Y.T

    if OPT == 'norm':
        return gs(X, row_vecs=False, norm=True)
    elif OPT == 'pad':
        return gs(X, row_vecs=False, norm=False)
    else:
        return X


def spm_hrf(RT, P=None, fMRI_T=16):
    """
    @RT - scan repeat time
    @P  - parameters of the response function (two gamma functions)

    defaults  (seconds)
    %	P[0] - Delay of Response (relative to onset)	    6
    %	P[1] - Delay of Undershoot (relative to onset)     16
    %	P[2] - Dispersion of Response			            1
    %	P[3] - Dispersion of Undershoot			            1
    %	P[4] - Ratio of Response to Undershoot		        6
    %	P[5] - Onset (seconds)				                0
    %	P[6] - Length of Kernel (seconds)		           32

    hrf  - hemodynamic response function
    P    - parameters of the response function
    """
    p = np.array([6, 16, 1, 1, 6, 0, 32], dtype=float)
    if P is not None:
        p[0:len(P)] = P
    _spm_Gpdf = lambda x, h, l: \
        np.exp(h * np.log(l) + (h - 1) * np.log(x) - (l * x) - gammaln(h))
    # modelled hemodynamic response function - {mixture of Gammas}
    dt = RT / float(fMRI_T)
    u = np.arange(0, int(p[6] / dt + 1)) - p[5] / dt
    with np.errstate(divide='ignore'):  # Known division-by-zero
        hrf = _spm_Gpdf(
            u, p[0] / p[2], dt / p[2]
        ) - _spm_Gpdf(
            u, p[1] / p[3], dt / p[3]
        ) / p[4]
    idx = np.arange(0, int((p[6] / RT) + 1)) * fMRI_T
    hrf = hrf[idx]
    hrf = np.nan_to_num(hrf)
    hrf = hrf / np.sum(hrf)
    return hrf


def spm_detrend(x, p=0):
    """
    Polynomial detrending over columns

    spm_detrend removes linear and nonlinear trends
    from column-wise data matrices.

    @x - data matrix
    @p - order of polynomial [default : 0]

    Returns:
    y - detrended data matrix
    """
    m, n = x.shape
    if (not m) or (not n):
        y = []
        return y

    if (not p):
        y = x - np.ones((m, 1), dtype='int') * x.mean(axis=0)
        return y

    G = np.zeros((m, p + 1))
    for i in range(0, p + 1):
        d = np.arange(1, m + 1) ** i
        G[:, i] = d.flatten(1)
    y = x - G.dot(np.linalg.pinv(G).dot(x))
    return y


def spm_write_vol(image_volume_info, image_voxels, image_name, file_type):
    """
    Writes an image volume to disk

    @image_volume_info - a structure containing image volume
     information (see spm_vol)
    @image_voxels - a one, two or three dimensional matrix
     containing the image voxels
    @image_name - name of the file to save the image in
    """
    if file_type == ".nii" or file_type == ".nii.gz":
        data = image_voxels
        affine = image_volume_info.affine
        image_volume_info = nib.Nifti1Image(data, affine)
        nib.save(image_volume_info, image_name + file_type)
    else:
        file_type = '.gii'
        data = image_voxels
        gi = nib.GiftiImage()
        gi.add_gifti_data_array(nib.gifti.GiftiDataArray(image_voxels))
        nib.gifti.giftiio.write(gi, image_name + file_type)



import numpy as np
from copy import deepcopy

from ...datatypes.misc.parameters import Parameters

class TimeSeries():
    """ 
    Deals with all the time-series artifacts that appear during the processing.

    These are the various time-series that are dealt with:
        1. BOLD              : (Raw BOLD Data)
        2. HRF               : (Hemodynamic Response Function)
        3. Preprocessed-BOLD : (Z-score and Passband filtered)
        4. Deconvolved-BOLD  : (Deconvolved using the HRF)

    Attributes:
        1. label         : Takes on one of the values as described above
        2. subject_index : index of the subject to which the current time-series belongs
        3. timeseries    : stores the time-series as a 2-dimensional numpy array
        4. shape         : shape of the time-series (voxels x time-slices)
        5. parameters    : rsHRF-parameters associated in retrieving this time-series
    """
    def __init__(self, label="",subject_index="",ts=np.array([]), para=Parameters()):
        self.label          = label
        self.subject_index  = subject_index
        self.timeseries     = deepcopy(ts)
        self.shape          = ts.shape
        self.parameters     = deepcopy(para) 
    
    # setters
    def set_ts(self, ts):
        self.timeseries     = deepcopy(ts) 
        self.shape          = self.timeseries.shape
    
    def set_parameters(self,para):
        self.parameters     = deepcopy(para) 
    
    def set_label(self,label):
        self.label          = label 
    
    def set_subject_index(self,subject_index):
        self.subject_index  = subject_index
    
    # getters
    def get_ts(self):
        return self.timeseries
    
    def get_subject_index(self):
        return self.subject_index
    
    def get_label(self):
        return self.label 
    
    def get_parameters(self):
        return self.parameters
    
    def get_shape(self):
        return self.shape 
    
    # misc.
    def get_info(self):
        """ Returns the information about the time-series in the form of a dictionary """
        dic                      = {}
        dic["Type"]              = self.label
        dic["Subject"]           = self.subject_index
        dic["Time Series Shape"] = self.shape
        dic["Parameters"]        = self.parameters.get_parameters()
        return dic

    def compareTimeSeries(self, ts):
        """ Compares another time-series with itself to determine if both are identical """
        raise NotImplementedError("This needs to be overridden in the child-classes")

    def save_info(self, name):
        """ Saves the information about the time-series in a .mat file """
        raise NotImplementedError("This needs to be overridden in the child-classes")




import pytest
from unittest import mock
import os
import math
import numpy   as np
import nibabel as nib
from scipy.special import gammaln
from ..spm_dep import spm

SHAPE = (10, 10, 10, 10)

def get_data(image_type):
    data = np.array(np.random.random(SHAPE), dtype=np.float32)
    mask = np.random.random(SHAPE[:3]) > 0.1
    if len(SHAPE) > 3:
        data[mask, :] = 0
    else:
        data[mask] = 0
    if image_type == 'nifti':
        data = nib.Nifti1Image(data, np.eye(4))
    else:
        data = nib.gifti.GiftiDataArray(data)
    return data

def test_spm_vol():
    test_file_1 = 'test.gii'
    test_file_2 = 'test.gii.gz'
    test_file_3 = 'test.nii'
    test_file_4 = 'test.nii.gz'
    test_files  = [test_file_1, test_file_2, test_file_3, test_file_4]
    with mock.patch('nibabel.load') as load_mock:
        for test_file in test_files:
            if 'nii' in test_file:
                load_mock.return_value = get_data('nifti')
            elif 'gii' in test_file:
                load_mock.return_value = get_data('gifti')
            v = spm.spm_vol(test_file)
            assert ('nii' in test_file or 'gii' in test_file)
            if 'nii' in test_file:
                assert type(v) == type(nib.Nifti1Image(np.asarray([]), np.eye(4)))
            elif 'gii' in test_file:
                assert type(v) == type(nib.gifti.GiftiDataArray(np.asarray([])))

def test_spm_read_vols():
    nifti = get_data('nifti')
    data = spm.spm_read_vols(nifti)
    assert type(data) == type(np.asarray([]))
    assert data.shape[0] == pow(10, 4)

def test_spm_orth():
    tests = [(3, 4), (7, 5), (4, 12), (13, 6), (11, 11)]
    for test in tests:
        X = np.random.random(test)
        Y = spm.spm_orth(X)
        assert type(Y) == type(X)
        assert Y.shape == X.shape

def test_spm_hrf():
    tests = [.5, 1, 2, 3, 4, 1.5, 2.5, 10]
    for test in tests:
        hrf = spm.spm_hrf(test)
        assert type(hrf) == type(np.asarray([]))
        assert len(hrf.shape) == 1
        assert hrf.size in [int(33/test) - 1, int(33/test), int(33/test) + 1] 

def test_spm_detrend():
    tests = [(3, 4), (7, 5), (4, 12), (13, 6), (11, 11)]
    for test in tests:
        X = np.random.random(test)
        Y = spm.spm_detrend(X)
        assert type(Y) == type(X)
        assert Y.shape == X.shape 
        Y = Y.T 
        Y_sum = np.sum(Y, axis=1)
        assert np.allclose(Y_sum, np.zeros(Y_sum.shape))

def test_spm_write_vol():
    test_file_1 = 'test.gii'
    test_file_2 = 'test.gii.gz'
    test_file_3 = 'test.nii'
    test_file_4 = 'test.nii.gz'
    test_files  = [test_file_1, test_file_2, test_file_3, test_file_4]
    with mock.patch('nibabel.load') as load_mock:
        for test_file in test_files:
            if 'nii' in test_file:
                load_mock.return_value = get_data('nifti')
            elif 'gii' in test_file:
                load_mock.return_value = get_data('gifti')
            v1 = spm.spm_vol(test_file)
            mask_data = np.zeros(SHAPE[:-1]).flatten(order='F')
            fname = test_file.split('.')[0]
            file_type = '.' + test_file.split('.', 1)[1]
            spm.spm_write_vol(v1, mask_data, fname, file_type)
            if 'gii' in file_type:
                file_type = '.gii'
            assert os.path.isfile(fname + file_type)
            os.remove(fname + file_type)


"""Base module variables."""
from os import path as op

with open(op.join(op.dirname(op.realpath(__file__)), "VERSION"), "r") as fh:
    __version__ = fh.read().strip('\n')

__packagename__ = 'rsHRF'
__url__ = 'https://github.com/BIDSapps/rsHRF'

DOWNLOAD_URL = (
    'https://github.com/BIDS-Apps/{name}/'.format(
        name=__packagename__))


import mpld3
import matplotlib
matplotlib.use("TkAgg")
import numpy as np
from matplotlib.figure                 import Figure
from tkinter                           import ttk, Toplevel, Canvas, TOP, BOTH, BOTTOM
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

class PlotterWindow():  
    def __init__(self):
        window = Toplevel()
        window.title("Screen")
        # get screen width and height
        screen_width       = window.winfo_screenwidth()
        screen_height      = window.winfo_screenheight()
        window.geometry("600x400+%d+%d" % (screen_width-600, screen_height/2))
        figure             = Figure(figsize=(5,5), dpi=100)
        self.numberOfPlots = 3
        self.ts            = [[] for i in range(0, self.numberOfPlots)]
        self.plot          = [figure.add_subplot(111) for i in range(0, self.numberOfPlots)]
        self.canvas        = FigureCanvasTkAgg(figure, window)

    def get_numberOfPlots(self):
        return self.numberOfPlots

    def makePlot(self, ts, val, num):
        for each in self.plot:
            each.clear()
        if val == 0:
            self.ts[num] = 0
        elif val == 1:
            self.ts[num] = ts
        for i in range(self.numberOfPlots):
            self.plot[i].plot(self.ts[i])
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(side=BOTTOM, fill=BOTH, expand=True)




import pytest
import numpy as np
from ..basis_functions import basis_functions

def test_fourier_bf():
    assert type(basis_functions.fourier_bf(np.random.random(5), {'estimation': 'fourier', 'order': 3})) == type(np.random.random(1))
    assert type(basis_functions.fourier_bf(np.random.random(5), {'estimation': 'fourier-hanning', 'order': 3})) == type(np.random.random(1))
    assert basis_functions.fourier_bf(np.random.random(5), {'estimation': 'fourier', 'order': 3}).shape == (5, 2*3 + 1)
    assert basis_functions.fourier_bf(np.random.random(5), {'estimation': 'fourier-hanning', 'order': 3}).shape == (5, 2*3 + 1)
    assert basis_functions.fourier_bf(np.random.random(15), {'estimation': 'fourier', 'order': 3}).shape == (15, 2*3 + 1)
    assert basis_functions.fourier_bf(np.random.random(15), {'estimation': 'fourier', 'order': 3}).shape == (15, 2*3 + 1)
    assert basis_functions.fourier_bf(np.random.random(7), {'estimation': 'fourier', 'order': 11}).shape == (7, 2*11 + 1)
    assert basis_functions.fourier_bf(np.random.random(7), {'estimation': 'fourier-hanning', 'order': 11}).shape == (7, 2*11 + 1)
    assert np.allclose(basis_functions.fourier_bf(np.zeros(5), {'estimation': 'fourier', 'order': 3}), np.asarray([[1, 0, 0, 0, 1, 1, 1] for i in range(5)]))
    assert np.allclose(basis_functions.fourier_bf(np.zeros(5), {'estimation': 'fourier-hanning', 'order': 3}), np.zeros((5, 2*3+1)))
    assert np.allclose(basis_functions.fourier_bf(np.ones(5), {'estimation': 'fourier', 'order': 3}), np.asarray([[1, 0, 0, 0, 1, 1, 1] for i in range(5)]))
    assert np.allclose(basis_functions.fourier_bf(np.ones(5), {'estimation': 'fourier-hanning', 'order': 3}), np.zeros((5, 2*3+1)))
    assert np.allclose(basis_functions.fourier_bf(np.full(5, np.pi/2), {'estimation': 'fourier', 'order': 3}), np.asarray([[ 1.,-0.43030122,0.77685322,-0.97220684,-0.90268536,0.62968173,-0.23412359] for i in range(5)]))
    assert np.allclose(basis_functions.fourier_bf(np.full(5, np.pi/2), {'estimation': 'fourier-hanning', 'order': 3}), np.asarray([[0.9513426809665357,-0.40936391340403033,0.7390536246669112,-0.9249018639367675,-0.8587631122906557,0.599043100699187,-0.222731764045654] for i in range(5)]))

def test_gamma_bf():
    assert type(basis_functions.gamma_bf(np.random.random(5), 7)) == type(np.random.random(1))
    assert basis_functions.gamma_bf(np.random.random(5), 7).shape == (5, 7)
    assert basis_functions.gamma_bf(np.random.random(6), 3).shape == (6, 3)
    assert basis_functions.gamma_bf(np.random.random(5), 17).shape == (5, 17)
    assert basis_functions.gamma_bf(np.random.random(15), 7).shape == (15, 7)
    assert basis_functions.gamma_bf(np.random.random(15), 11).shape == (15, 11)
    assert basis_functions.gamma_bf(np.random.random(15), 27).shape == (15, 27)
    assert np.allclose(basis_functions.gamma_bf(np.zeros(5), 7), np.zeros((5, 7)))
    assert np.allclose(basis_functions.gamma_bf(np.ones(5), 3), np.asarray([[6.13132402e-02, 7.29919526e-05, 2.81323432e-13] for i in range(5)]))


import numpy as np
import warnings

warnings.filterwarnings("ignore")

def nextpow2(n):
    return np.ceil(np.log2(np.abs(n))).astype('long')


def rest_nextpow2_one35(n):
    if type(n) is np.ndarray:
        if(np.amax(n.shape)) > 1:
            n = np.amax(n.shape)
    if n < 16:
        Result = 2 ** nextpow2(n)
        return Result
    limit = nextpow2(n)
    tbl = np.arange((2 ** (limit - 1)), ((2 ** limit) + 1))
    tbl = np.extract(tbl >= n, tbl)
    for x in range(0, np.amax(tbl.shape)):
        Result = tbl[x]
        f, p = np.frexp(Result)
        if f.size != 0 and f == 0.5:
            return Result
        if np.remainder(Result, 3 * 5) == 0:
            y = Result / (3 * 5)
            f, p = np.frexp(y)
            if f.size != 0 and f == 0.5:
                return Result
        if np.remainder(Result, 3) == 0:
            y = Result / 3
            f, p = np.frexp(y)
            if f.size != 0 and f == 0.5:
                return Result
        if np.remainder(Result, 5) == 0:
            y = Result / 5
            f, p = np.frexp(y)
            if f.size != 0 and f == 0.5:
                return Result
    Result = np.nan
    return Result


def rest_IdealFilter(Data, SamplePeriod, Band):
    sampleFreq = 1 / SamplePeriod
    sampleLength = Data.shape[0]
    paddedLength = rest_nextpow2_one35(sampleLength)
    LowCutoff_HighPass = Band[0]
    HighCutoff_LowPass = Band[1]
    if (LowCutoff_HighPass >= sampleFreq / 2):
        idxLowCutoff_HighPass = int(paddedLength / 2 + 1)
    else:
        idxLowCutoff_HighPass = int(
            np.ceil(LowCutoff_HighPass * paddedLength * SamplePeriod + 1)
        )

    if (HighCutoff_LowPass >= sampleFreq / 2) or (HighCutoff_LowPass == 0):
        idxHighCutoff_LowPass = int(paddedLength / 2 + 1)
    else:
        idxHighCutoff_LowPass = int(
            np.fix(HighCutoff_LowPass * paddedLength * SamplePeriod + 1)
        )

    FrequencyMask = np.zeros((paddedLength, 1))
    FrequencyMask[idxLowCutoff_HighPass - 1:idxHighCutoff_LowPass, 0] = 1
    FrequencyMask[paddedLength - idxLowCutoff_HighPass + 1:
                  paddedLength - idxHighCutoff_LowPass:-1, 0] = 1
    FrequencySetZero_Index = np.nonzero(FrequencyMask == 0)

    Data = Data - np.tile(np.mean(Data, axis=0), (Data.shape[0], 1))
    Data = np.concatenate(
        (Data, np.zeros((paddedLength - sampleLength, Data.shape[1]))), axis=0
    )
    Data = np.fft.fft(Data, axis=0)
    Data[FrequencySetZero_Index, :] = 0
    Data = np.fft.ifft(Data, axis=0)
    Data_Filtered = Data[0:sampleLength, :]
    return Data_Filtered



"""Utilities to handle BIDS inputs."""
import os
import sys
import json
from pathlib import Path

def write_derivative_description(bids_dir, deriv_dir):
    from ..__about__ import __version__, DOWNLOAD_URL

    print(__version__)
    bids_dir = Path(bids_dir)
    deriv_dir = Path(deriv_dir)
    desc = {
        'Name': 'rsHRF - retrieve the haemodynamic response function from resting state fMRI data',
        'BIDSVersion': '1.4.0',
        'DatasetType': 'derivative',
        'GeneratedBy': [{
            'Name': 'rsHRF',
            'Version': __version__,
            'CodeURL': DOWNLOAD_URL,
        }],
        'HowToAcknowledge':
            'Please cite our paper (under review). '
            ,
    }

    # Keys that can only be set by environment
    if 'RSHRF_DOCKER_TAG' in os.environ:
        desc['GeneratedBy'][0]['Container'] = {
            "Type": "docker",
            "Tag": f"bids/rshrf:{os.environ['FMRIPREP_DOCKER_TAG']}"
        }
    if 'RSHRF_SINGULARITY_URL' in os.environ:
        desc['GeneratedBy'][0]['Container'] = {
            "Type": "singularity",
            "URI": os.getenv('FMRIPREP_SINGULARITY_URL')
        }

    # Keys deriving from source dataset
    orig_desc = {}
    fname = bids_dir / 'dataset_description.json'
    if fname.exists():
        orig_desc = json.loads(fname.read_text())

    if 'DatasetDOI' in orig_desc:
        desc['SourceDatasets'] = [{
            'URL': f'https://doi.org/{orig_desc["DatasetDOI"]}',
            'DOI': orig_desc['DatasetDOI']
        }]
    if 'License' in orig_desc:
        desc['License'] = orig_desc['License']

    Path.write_text(deriv_dir / 'dataset_description.json', json.dumps(desc, indent=4))


import pyyawt
import numpy as np 
from   rsHRF.processing import knee 

def rsHRF_iterative_wiener_deconv(y, h, Iterations=1000):
    N            = y.shape[0]
    nh           = max(h.shape)
    h            = np.append(h, np.zeros((N - nh, 1)))
    H            = np.fft.fft(h, axis=0)
    Y            = np.fft.fft(y, axis=0)
    [c ,l]       = pyyawt.wavedec(abs(y), 1, 'db2')
    sigma        = pyyawt.wnoisest(c, l, 1)
    Phh          = np.square(abs(H))
    sqrdtempnorm = ((((np.linalg.norm(y-np.mean(y), 2)**2) - (N-1)*(sigma**2)))/(np.linalg.norm(h,1))**2)
    Nf           = (sigma**2)*N
    tempreg      = Nf/sqrdtempnorm
    Pxx0         = np.square(abs(np.multiply(Y, (np.divide(np.conj(H), (Phh + N*tempreg))))))
    Pxx          = Pxx0
    Sf           = Pxx.reshape(-1, 1)
    for i in range(0, Iterations):
        M           = np.divide(np.multiply(np.multiply(np.conjugate(H), Pxx), Y), np.add(np.multiply(np.square(abs(H)), Pxx), Nf))
        PxxY        = np.divide(np.multiply(Pxx, Nf), np.add(np.multiply(np.square(abs(H)), Pxx), Nf))
        Pxx         = np.add(PxxY, np.square(abs(M)))
        Sf          = np.concatenate((Sf, Pxx.reshape(-1, 1)), axis = 1)
    dSf             = np.diff(Sf, 1, 1)
    dSfmse          = np.mean(np.square(dSf), axis=1)
    _, idx          = knee.knee_pt(dSfmse)
    idm             = np.argmin(dSfmse)
    ratio           = np.abs(dSfmse[idx] - dSfmse[idm])/(np.abs(np.max(dSfmse) - np.min(dSfmse)))
    if ratio > 0.5:
        id0 = idm 
    else:
        id0 = idx 
    Pxx             = Sf[:,id0+1]
    WienerFilterEst = np.divide(np.multiply(np.conj(H), Pxx), np.add(np.multiply(np.square(abs(H)), Pxx), Nf))
    return np.real(np.fft.ifft(np.multiply(WienerFilterEst, Y)))


import os
import shutil
import tempfile
import numpy as np
from scipy        import stats
from scipy.sparse import lil_matrix
from joblib       import load, dump
from joblib       import Parallel, delayed
from rsHRF        import processing, sFIR
from ..processing import knee

import warnings
warnings.filterwarnings("ignore")

"""
HRF ESTIMATION
"""

def compute_hrf(bold_sig, para, temporal_mask, p_jobs, bf = None):
    para['temporal_mask'] = temporal_mask
    N, nvar = bold_sig.shape
    folder = tempfile.mkdtemp()
    data_folder = os.path.join(folder, 'data')
    dump(bold_sig, data_folder)
    data = load(data_folder, mmap_mode='r')
    results = Parallel(n_jobs=p_jobs)(delayed(estimate_hrf)(data, i, para,
                                  N, bf) for i in range(nvar))
    beta_hrf, event_bold = zip(*results)
    try:
        shutil.rmtree(folder)
    except:
        print("Failed to delete: " + folder)
    return np.array(beta_hrf).T, np.array(event_bold)

def estimate_hrf(bold_sig, i, para, N, bf = None):
    """
    Estimate HRF
    """
    dat = bold_sig[:, i]
    localK = para['localK']
    if para['estimation'] == 'sFIR' or para['estimation'] == 'FIR':
        #Estimate HRF for the sFIR or FIR basis functions
        if np.count_nonzero(para['thr']) == 1:
            para['thr'] = np.array([para['thr'], np.inf])
        thr = para['thr'] #Thr is a vector for (s)FIR
        u = wgr_BOLD_event_vector(N, dat, thr, localK, para['temporal_mask'])
        u = u.toarray().flatten('C').ravel().nonzero()[0]
        beta_hrf, event_bold = sFIR.smooth_fir.wgr_FIR_estimation_HRF(u, dat, para, N)
    else:
        thr = [para['thr']] #Thr is a scalar for the basis functions
        u0 = wgr_BOLD_event_vector(N, dat, thr, localK, para['temporal_mask'])
        u = np.append(u0.toarray(), np.zeros((para['T'] - 1, N)), axis=0)
        u = np.reshape(u, (1, - 1), order='F')
        beta_hrf = wgr_hrf_fit(dat, para, u, bf)
        u = u0.toarray()[0].nonzero()[0]
    return beta_hrf, u

def wgr_onset_design(u, bf, T, T0, nscans):
    """
    @u - BOLD event vector (microtime).
    @bf - basis set matrix
    @T - microtime resolution (number of time bins per scan)
    @T0 - microtime onset (reference time bin, see slice timing)
    """
    ind = np.arange(0, max(u.shape))
    X = np.empty((0, len(ind)))
    for p in range(bf.shape[1]):
        x = np.convolve(u, bf[:, p])
        x = x[ind]
        X = np.append(X, [x], axis=0)
    X = X.T
    """
    Resample regressors at acquisition times
    """
    X = X[(np.arange(0, nscans) * T) + (T0 - 1), :]
    return X


def wgr_glm_estimation(dat, u, bf, T, T0, AR_lag):
    """
    @u - BOLD event vector (microtime).
    """
    nscans = dat.shape[0]
    x = wgr_onset_design(u, bf, T, T0, nscans)
    X = np.append(x, np.ones((nscans, 1)), axis=1)
    res_sum, Beta = sFIR.smooth_fir.wgr_glsco(X, dat, AR_lag=AR_lag)
    return np.real(res_sum), Beta

def wgr_hrf_fit(dat, xBF, u, bf):
    """
    @u    - BOLD event vector (microtime).
    @nlag - time lag from neural event to BOLD event
    """
    lag = xBF['lag']
    AR_lag = xBF['AR_lag']
    nlag = len(lag)
    erm = np.zeros((1, nlag))
    beta = np.zeros((bf.shape[1] + 1, nlag))
    for i in range(nlag):
        u_lag = np.append(u[0, lag[i]:], np.zeros((1, lag[i]))).T
        erm[0, i], beta[:, i] = \
            wgr_glm_estimation(dat, u_lag, bf, xBF['T'], xBF['T0'], AR_lag)
    x, idx = knee.knee_pt(np.ravel(erm))
    if idx == nlag-1:
        idx = idx - 1
    beta_hrf = beta[:, idx+1]
    beta_hrf = np.append(beta_hrf, lag[idx+1])
    return beta_hrf

def wgr_BOLD_event_vector(N, matrix, thr, k, temporal_mask):
    """
    Detect BOLD event.
    event > thr & event < 3.1
    """
    data = lil_matrix((1, N))
    matrix = matrix[:, np.newaxis]
    matrix = np.nan_to_num(matrix)
    if 0 in np.array(temporal_mask).shape:
        matrix = stats.zscore(matrix, ddof=1)
        for t in range(1 + k, N - k + 1):
            if matrix[t - 1, 0] > thr[0] and \
                    np.all(matrix[t - k - 1:t - 1, 0] < matrix[t - 1, 0]) and \
                    np.all(matrix[t - 1, 0] > matrix[t:t + k, 0]):
                data[0, t - 1] = 1
    else:
        tmp = temporal_mask
        for i in range(len(temporal_mask)):
            if tmp[i] == 1:
                temporal_mask[i] = i
        datm = np.mean(matrix[temporal_mask])
        datstd = np.std(matrix[temporal_mask])
        if datstd == 0: datstd = 1
        matrix = (matrix - datm)/datstd
        for t in range(1 + k, N - k + 1):
            if tmp[t-1]:
                if matrix[t - 1, 0] > thr[0] and \
                        np.all(matrix[t - k - 1:t - 1, 0] < matrix[t - 1, 0]) and \
                        np.all(matrix[t - 1, 0] > matrix[t:t + k, 0]):
                    data[0, t - 1] = 1.
    return data



import numpy as np
from ..spm_dep    import spm

import warnings
warnings.filterwarnings("ignore")

def wgr_spm_get_canonhrf(xBF):
    dt     = xBF['dt']
    fMRI_T = xBF['T']
    p      = np.array([6, 16, 1, 1, 6, 0, 32], dtype=float)
    p[len(p) - 1] = xBF['len']
    bf = spm.spm_hrf(dt, p, fMRI_T)
    bf = bf[:, np.newaxis]
    # time-derivative
    if xBF['TD_DD']:
        dp   = 1
        p[5] = p[5] + dp
        D    = (bf[:, 0] - spm.spm_hrf(dt, p, fMRI_T)) / dp
        D    = D[:, np.newaxis]
        bf   = np.append(bf, D, axis=1)
        p[5] = p[5] - dp
        # dispersion-derivative
        if xBF['TD_DD'] == 2:
            dp   = 0.01
            p[2] = p[2] + dp
            D    = (bf[:, 0] - spm.spm_hrf(dt, p, fMRI_T)) / dp
            D    = D[:, np.newaxis]
            bf   = np.append(bf, D, axis=1)
    return bf



from .plotterWindow import PlotterWindow
from tkinter import Toplevel, OptionMenu, StringVar, _setit, Entry, IntVar, Checkbutton

class PlotterOptionsWindow():
    def __init__(self):
        window = Toplevel()
        window.title("Plotting Menu")
        # get screen width and height
        screen_width  = window.winfo_screenwidth()
        screen_height = window.winfo_screenheight()
        window.geometry("650x200+%d+%d" % (screen_width/2.95, 100 + screen_height/2))
        self.plotterScreen = PlotterWindow()
        
        # can have 3 different plots at once
        self.numberOfPlots = self.plotterScreen.get_numberOfPlots()
        self.plot          = [StringVar() for i in range(self.numberOfPlots)]
        self.plotVal       = [IntVar() for i in range(self.numberOfPlots)]
        self.plotValStore  = [0 for i in range(self.numberOfPlots)]
        self.plotables     = []
        self.options       = ["None"]
        for each in self.plot:
            each.set("None")

        def plotTS():
            for i in range(0, self.numberOfPlots):
                if self.plotVal[i].get() != self.plotValStore[i]:
                    self.plotValStore[i]  = self.plotVal[i].get()
                    if self.plotValStore[i] == 0:
                        self.plotterScreen.makePlot(None, 0, i)
                        return
                    for each in self.plotables:
                        if each[0] == self.plot[i].get():
                            self.plotterScreen.makePlot(each[1][:,int(self.selectVoxel[i].get())], 1, i)
                            return 
                    self.plotterScreen.makePlot(None, 0, i)
                    return   
                
        self.plotSelectDropDown = [OptionMenu(window, self.plot[i], *self.options) for i in range(self.numberOfPlots) for i in range(self.numberOfPlots)]
        self.selectVoxel        = [Entry(window, width=10) for i in range(self.numberOfPlots) for i in range(self.numberOfPlots)]
        self.plotButton         = [Checkbutton(window, text=" Plot TS " + str(i), variable=self.plotVal[i], command=plotTS) for i in range(self.numberOfPlots)]

        for each in self.selectVoxel:
            each.insert(0, 0)

        for i in range(0, self.numberOfPlots):
            self.plotSelectDropDown[i].grid(row=0,column=i,padx=(5,5),pady=(5,5))
            self.selectVoxel[i].grid       (row=1,column=i,padx=(5,5),pady=(5,5))
            self.plotButton[i].grid        (row=2,column=i,padx=(5,5),pady=(5,5))

    def updatePlots(self, plotables):
        temp_names = []
        for each in self.plotables:
            temp_names.append(each[0])
        for each in plotables:
            if each[0] not in temp_names:
                self.plotables.append(each)
        self.options = ["None"]
        for each in self.plotables:
            self.options.append(each[0])
        self.options[1:] = sorted(self.options[1:])
        self.updateWidgets()

    def updateWidgets(self):

        if len(self.options) == 0:
            self.options.append("None")

        for each in self.plotSelectDropDown:
            each['menu'].delete(0, 'end')

        for choice in self.options:
            for i in range(0, self.numberOfPlots):
                self.plotSelectDropDown[i]['menu'].add_command(label=choice, command=_setit(self.plot[i], choice))

    def get_plotables(self):
        return self.plotables



import numpy as np
import warnings

warnings.filterwarnings("ignore")

def wgr_get_parameters(hdrf, dt):
    """
    Find Model Parameters
    h - Height
    p - Time to peak (in units of dt where dt = TR/para.T)
    w - Width at half-peak
    """
    param = np.zeros((3, 1))

    if(np.any(hdrf)):
        n = np.fix(np.amax(hdrf.shape) * 0.8)

        p = np.argmax(np.absolute(hdrf[np.arange(0, n, dtype='int')]))
        h = hdrf[p]

        if h > 0:
            v = (hdrf >= (h / 2))
        else:
            v = (hdrf <= (h / 2))
        v = v.astype(int)
        b = np.argmin(np.diff(v))
        v[b + 1:] = 0
        w = np.sum(v)

        cnt = p - 1
        g = hdrf[1:] - hdrf[0:-1]

        while cnt > 0 and np.abs(g[cnt]) < 0.001:
            h = hdrf[cnt - 1]
            p = cnt
            cnt = cnt - 1

        param[0] = h
        param[1] = (p + 1) * dt
        param[2] = w * dt

    else:
        print('.')
    return param.ravel()



from ..misc.status        import Status
from datetime             import datetime
from tkinter              import Toplevel, BOTH, INSERT
from tkinter.scrolledtext import ScrolledText

class LoggingWindow():
    def __init__(self):
        window = Toplevel(bg='white')
        window.title("Logger")
        # get screen width and height
        screen_width  = window.winfo_screenwidth()
        screen_height = window.winfo_screenheight()
        window.geometry("600x400+%d+%d" % (screen_width-600, 0))
        self.lineNum  = 1
        self.text     = ScrolledText(window, background='black')
        self.text.pack(fill=BOTH, expand = 1)

    def putLog(self, plotables=None, data_info=None, status=None):
        # writing current time
        current_time = datetime.now().strftime("%H:%M:%S")

        if status is not None:
            if status.get_state():
                message = status.get_info()
                self.text.insert(INSERT, \
                    ">> " + current_time + "\t SUCCESS " + message + "\n"
                )
                self.text.tag_add("success", str(self.lineNum) + ".12", str(self.lineNum) + ".20")
                self.text.tag_config("success", foreground="green")
                self.text.tag_add("success message", str(self.lineNum) + ".20", str(self.lineNum) + "." + str(21 + len(message)))
                self.text.tag_config("success message", foreground="LightBlue1")
            else:
                message = status.get_error()
                self.text.insert(INSERT, \
                    ">> " + current_time + "\t ERROR " + message + "\n"
                )
                self.text.tag_add("error", str(self.lineNum) + ".12", str(self.lineNum) + ".18")
                self.text.tag_config("error", foreground="red")
                self.text.tag_add("error message", str(self.lineNum) + ".18", str(self.lineNum) + "." + str(19 + len(message)))
                self.text.tag_config("error message", foreground="yellow")    
            self.text.tag_add("time", str(self.lineNum) + ".03", str(self.lineNum) + ".11")
            self.text.tag_config("time", foreground="PeachPuff2")
            self.lineNum += 1

            dic = status.get_dic()
            for each in dic.keys():
                self.text.insert(INSERT, \
                    ">> " + current_time + "\t" + each + str(dic[each]) + "\n"
                )
                self.text.tag_add("time", str(self.lineNum) + ".03", str(self.lineNum) + ".11")
                self.text.tag_config("time", foreground="PeachPuff2")
                self.text.tag_add("dic_label", str(self.lineNum) + ".12", str(self.lineNum) + "." + str(12+len(each)))
                self.text.tag_config("dic_label", foreground="azure")
                self.lineNum += 1
            self.text.insert(INSERT, "\n")
            self.lineNum += 1
            return

        elif plotables is not None:
            self.text.insert(INSERT, \
                ">> " + current_time + "\t Time Series' Information \n\n" + \
                    "TYPE" + "\t\t\t\t\t\t LENGTH" + "\t\t NUMBER OF VOXELS \n\n"
                )
            self.text.tag_add("labels", str(self.lineNum) + ".12", str(self.lineNum) + "." + str(37))
            self.text.tag_config("labels", foreground="PaleTurquoise1")
            self.text.tag_add("titles", str(self.lineNum+2) + ".00", str(self.lineNum+2) + "." + str(37))
            self.text.tag_config("titles", foreground="azure")
            self.text.tag_add("time", str(self.lineNum) + ".03", str(self.lineNum) + ".11")
            self.text.tag_config("time", foreground="PeachPuff2")
            for each in plotables:
                self.text.insert(INSERT, \
                    each[0] + "\t\t\t\t\t\t " + str(each[1].shape[0]) + "\t\t " + str(each[1].shape[1]) + "\n"
                    )
                self.lineNum += 1
            self.text.insert(INSERT, "\n")
            self.lineNum += 5
            return
        
        elif data_info is not None:
            self.text.insert(INSERT, \
                ">> " + current_time + "\t Information \n\n")
            self.text.tag_add("labels", str(self.lineNum) + ".12", str(self.lineNum) + "." + str(37))
            self.text.tag_config("labels", foreground="PaleTurquoise1")
            self.text.tag_add("time", str(self.lineNum) + ".03", str(self.lineNum) + ".11")
            self.text.tag_config("time", foreground="PeachPuff2")
            self.lineNum += 2
            if data_info["Type"] == "BOLD":
                self.text.insert(INSERT, \
                    "   :- " + "Type of Data: BOLD Time Series\n" + 
                    "   :- " + "Subject: " + data_info["Subject"] + "\n"
                    "   :- " + "Input File: " + data_info["Input File"] + "\n"
                    )
                self.lineNum += 1
            elif data_info["Type"] == "Preprocessed-BOLD":
                self.text.insert(INSERT, \
                    "   :- " + "Type of Data: Preprocessed-BOLD Time Series\n" + 
                    "   :- " + "Subject: " + data_info["Subject"] + "\n" +
                    "   :- " + "Mask File: " + data_info["Mask File"] + "\n" + 
                    "   :- " + "Associated Raw BOLD: " + data_info["Associated Raw BOLD"] + "\n"
                    )
                self.lineNum += 2
            elif data_info["Type"] == "HRF":
                self.text.insert(INSERT, \
                    "   :- " + "Type of Data: Hemodynamic Response Function Time Series\n" + 
                    "   :- " + "Associated BOLD: " + data_info["Associated BOLD"] + "\n"
                    )
            elif data_info["Type"] == "Deconvolved-BOLD":
                self.text.insert(INSERT, \
                    "   :- " + "Type of Data: Deconvolved BOLD Time Series\n" + 
                    "   :- " + "Associated HRF: " + data_info["Associated HRF"] + "\n"
                    )
            self.text.tag_config("type", foreground="azure")
            self.text.tag_config("line2", foreground="azure")
            self.lineNum += 2 
            self.text.insert(INSERT, \
                    "   :- " + "\t\t\tAssociated Parameters\n" 
                    )
            self.text.tag_add("ap", str(self.lineNum) + ".03", str(self.lineNum) + ".30")
            self.text.tag_config("ap", foreground="Khaki1")
            self.lineNum += 1
            for each in data_info["Parameters"].keys():
                self.text.insert(INSERT, \
                    "\t\t" + each + "\t\t\t=\t\t" + str(data_info["Parameters"][each]) + "\n")
                self.text.tag_add("pv", str(self.lineNum) + ".00", str(self.lineNum) + "." + str(6+len(each)))
                self.text.tag_config("pv", foreground="LemonChiffon2")
                self.lineNum +=1
            self.text.insert(INSERT, "\n")
            self.lineNum += 1
            return


import sys
import numpy   as np
import os.path as op
import json
from argparse      import ArgumentParser
from bids.layout import BIDSLayout
from pathlib import Path
from rsHRF      import spm_dep, fourD_rsHRF, utils

import warnings
warnings.filterwarnings("ignore")

with open(op.join(op.dirname(op.realpath(__file__)), "VERSION"), "r") as fh:
    __version__ = fh.read().strip('\n')

def get_parser():
    parser = ArgumentParser(description='retrieves the onsets of pseudo-events triggering a '
                                        'haemodynamic response from resting state fMRI BOLD '
                                        'voxel-wise signal')

    group_input = parser.add_mutually_exclusive_group(required=True)

    group_input.add_argument('--ts', action='store', type=op.abspath,
                             help='the absolute path to a single data file')

    group_input.add_argument('--input_file', action='store', type=op.abspath,
                             help='the absolute path to a single data file')

    group_input.add_argument('--bids_dir', nargs='?', action='store', type=op.abspath,
                             help='the root folder of a BIDS valid dataset '
                                  '(sub-XXXXX folders should be found at the '
                                  'top level in this folder).')

    group_input.add_argument('--GUI', action='store_true',
                            help='to execute the toolbox in GUI mode')

    parser.add_argument('--output_dir', action='store', type=op.abspath,
                        help='the output path for the outcomes of processing')

    parser.add_argument('--n_jobs', action='store', type=int, default=-1,
                        help='the number of parallel processing elements')

    parser.add_argument('-V', '--version', action='version', version='rsHRF version {}'.format(__version__))

    parser.add_argument('--analysis_level', help='Level of the analysis that will be performed. '
                        'Multiple participant level analyses can be run independently '
                        '(in parallel) using the same output_dir.', choices=['participant'], nargs='?')

    parser.add_argument('--participant_label',
                        help='The label(s) of the participant(s) that should be analyzed. The label '
                             'corresponds to sub-<participant_label> from the BIDS spec '
                             '(so it does not include "sub-"). If this parameter is not '
                             'provided all subjects should be analyzed. Multiple '
                             'participants can be specified with a space separated list.',
                        nargs="+")
   
    parser.add_argument('--bids_filter_file', action='store', type=op.abspath,
                        help='a JSON file describing custom BIDS input filters using PyBIDS. '
                             'For further details, please check out http://bids-apps.neuroimaging.io/rsHRF/')
    
    group_mask = parser.add_mutually_exclusive_group(required=False)

    group_mask.add_argument('--atlas', action='store', type=op.abspath,
                            help='the absolute path to a single atlas file')

    group_mask.add_argument('--brainmask', action='store_true',
                            help='to enable the use of mask files present in the BIDS '
                                 'directory itself')

    group_para = parser.add_argument_group('Parameters')

    group_para.add_argument('--estimation', action='store',
                            choices=['canon2dd', 'sFIR', 'FIR', 'fourier', 'hanning', 'gamma'],
                            help='Choose the estimation procedure from '
                                 'canon2dd (canonical shape with 2 derivatives), '
                                 'sFIR (smoothed Finite Impulse Response), '
                                 'FIR (Finite Impulse Response), '
                                 'fourier (Fourier Basis Set), '
                                 'hanning (Fourier Basis w Hanning), '
                                 'gamma (Gamma Basis Set)')

    group_para.add_argument('--passband', action='store', type=float, nargs=2, metavar=('LOW_FREQ','HIGH_FREQ'),
                            default=[0.01, 0.08],
                            help='set intervals for bandpass filter, default is 0.01 - 0.08')

    group_para.add_argument('--passband_deconvolve', action='store', type=float, nargs=2, metavar=('LOW_FREQ', 'HIGH_FREQ'),
                            default=[0.0, sys.float_info.max],
                            help='set intervals for bandpass filter (used while deconvolving BOLD), default is no-filtering')

    group_para.add_argument('-TR', action='store', type=float, default=-1,
                            help='set TR parameter')

    group_para.add_argument('-T', action='store', type=int, default=3,
                            help='set T parameter')

    group_para.add_argument('-T0', action='store', type=int, default=1,
                            help='set T0 parameter')

    group_para.add_argument('-TD_DD', action='store', type=int, default=2,
                            help='set TD_DD parameter')

    group_para.add_argument('-AR_lag', action='store', type=int, default=1,
                            help='set AR_lag parameter')

    group_para.add_argument('--thr', action='store', type=float, default=1,
                            help='set thr parameter')

    group_para.add_argument('--temporal_mask', action='store', type=op.abspath,
                        help='the path for the (temporal) mask file.\n The mask file should be a ".dat" file, consisting of a binary string of the same length as the signal')

    group_para.add_argument('--order', action='store', type=int, default=3,
                            help='set the number of basis vectors')

    group_para.add_argument('--len', action='store', type=int, default=24,
                            help='set len parameter')

    group_para.add_argument('--min_onset_search', action='store', type=int, default=4,
                            help='set min_onset_search parameter')

    group_para.add_argument('--max_onset_search', action='store', type=int, default=8,
                            help='set max_onset_search parameter')

    group_para.add_argument('--localK', action='store', type=int,
                            help='set localK')

    group_para.add_argument('--wiener', action='store_true',
                            help='to perform iterative wiener deconvolution')

    return parser


def run_rsHRF():
    parser     = get_parser()
    args       = parser.parse_args()
    arg_groups = {}
    for group in parser._action_groups:
        group_dict              = {a.dest: getattr(args, a.dest, None) for a in group._group_actions }
        arg_groups[group.title] = group_dict
    para          = arg_groups['Parameters']
    nargs         = len(sys.argv)
    temporal_mask = []

    if (not args.GUI) and (args.output_dir is None):
        parser.error('--output_dir is required when executing in command-line interface')

    if (not args.GUI) and (args.estimation is None):
        parser.error('--estimation rule is required when executing in command-line interface')

    if (args.GUI):
        if (nargs == 2):
            try:
                from .rsHRF_GUI import run
                run.run()
            except ModuleNotFoundError:
                parser.error('--GUI should not be used inside a Docker container')
        else:
            parser.error('--no other arguments should be supplied with --GUI')

    if (args.input_file is not None or args.ts is not None) and args.analysis_level:
        parser.error('analysis_level cannot be used with --input_file or --ts, do not supply it')

    if (args.input_file is not None or args.ts is not None) and args.participant_label:
        parser.error('participant_labels are not to be used with --input_file or --ts, do not supply it')

    if args.input_file is not None and args.brainmask:
        parser.error('--brainmask cannot be used with --input_file, use --atlas instead')

    if args.ts is not None and (args.brainmask or args.atlas):
        parser.error('--atlas or --brainmask cannot be used with --ts, do not supply it')

    if args.bids_dir is not None and not (args.brainmask or args.atlas):
        parser.error('--atlas or --brainmask needs to be supplied with --bids_dir')

    if args.bids_dir is not None and not args.analysis_level:
        parser.error('analysis_level needs to be supplied with bids_dir, choices=[participant]')

    if args.input_file is not None and (not args.input_file.endswith(('.nii', '.nii.gz', '.gii', '.gii.gz'))):
        parser.error('--input_file should end with .gii, .gii.gz, .nii or .nii.gz')

    if args.atlas is not None and (not args.atlas.endswith(('.nii', '.nii.gz','.gii', '.gii.gz'))):
        parser.error('--atlas should end with .gii, .gii.gz, .nii or .nii.gz')

    if args.ts is not None and (not args.ts.endswith(('.txt'))):
        parser.error('--ts file should end with .txt')

    if args.temporal_mask is not None and (not args.temporal_mask.endswith(('.dat'))):
        parser.error('--temporal_mask ile should end with ".dat"')

    if args.temporal_mask is not None:
        f = open(args.temporal_mask,'r')
        for line in f:
            for each in line:
                if each in ['0','1']:
                    temporal_mask.append(int(each))

    if args.estimation == 'sFIR' or args.estimation == 'FIR':
        para['T'] = 1

    if args.ts is not None:
        file_type = op.splitext(args.ts)
        if para['TR'] <= 0:
            parser.error('Please supply a valid TR using -TR argument')
        else:
            TR = para['TR']
        para['dt'] = para['TR'] / para['T']
        para['lag'] = np.arange(np.fix(para['min_onset_search'] / para['dt']),
                                np.fix(para['max_onset_search'] / para['dt']) + 1,
                                dtype='int')
        fourD_rsHRF.demo_rsHRF(args.ts, None, args.output_dir, para, args.n_jobs, file_type, mode='time-series', temporal_mask=temporal_mask, wiener=args.wiener)

    if args.input_file is not None:
        if args.atlas is not None:
            if (args.input_file.endswith(('.nii', '.nii.gz')) and args.atlas.endswith(('.gii', '.gii.gz'))) or (args.input_file.endswith(('.gii', '.gii.gz')) and args.atlas.endswith(('.nii', '.nii.gz'))):
                parser.error('--atlas and input_file should be of the same type [NIfTI or GIfTI]')

        # carry analysis with input_file and atlas
        file_type = op.splitext(args.input_file)
        if file_type[-1] == ".gz":
            file_type = op.splitext(file_type[-2])[-1] + file_type[-1]
        else:
            file_type = file_type[-1]
        if ".nii" in file_type:
            TR = (spm_dep.spm.spm_vol(args.input_file).header.get_zooms())[-1]
        else:
            if para['TR'] == -1:
                parser.error('Please supply a valid TR using -TR argument')
            else:
                TR = para['TR']
        if TR <= 0:
            if para['TR'] <= 0:
                parser.error('Please supply a valid TR using -TR argument')
        else:
            if para['TR'] == -1:
                para['TR'] = TR
            elif para['TR'] <= 0:
                print('Invalid TR supplied, using implicit TR: {0}'.format(TR))
                para['TR'] = TR
        para['dt'] = para['TR'] / para['T']
        para['lag'] = np.arange(np.fix(para['min_onset_search'] / para['dt']),
                                np.fix(para['max_onset_search'] / para['dt']) + 1,
                                dtype='int')
        fourD_rsHRF.demo_rsHRF(args.input_file, args.atlas, args.output_dir, para, args.n_jobs, file_type, mode='input', temporal_mask=temporal_mask, wiener=args.wiener)

    if args.bids_dir is not None:
        utils.bids.write_derivative_description(args.bids_dir, args.output_dir)
        bids_dir = Path(args.bids_dir)
        fname = bids_dir / 'dataset_description.json'
        
        if fname.exists():
                desc = json.loads(Path(fname).read_text())
                if 'DataType' in desc :
                    if desc['DataType'] != 'derivative':
                        parser.error('Input data is not a derivative dataset'
                                     ' (DataType in dataset_description.json is not equal to "derivative")')
                                       
                else :
                    parser.error('DataType is not defined in the dataset_description.json file. Please make sure DataType is defined. '
                                 'Information on the dataset_description.json file can be found online '
                                 '(https://bids-specification.readthedocs.io/en/stable/03-modality-agnostic-files.html'
                                 '#derived-dataset-and-pipeline-description)')
        else :
            parser.error('Could not find dataset_description.json file. Please make sure the BIDS data '
                         'structure is present and correct. Datasets can be validated online '
                         'using the BIDS Validator (http://incf.github.io/bids-validator/).')
    
        
    if args.bids_dir is not None and args.atlas is not None:
        # carry analysis with bids_dir and 1 atlas
        layout = BIDSLayout(args.bids_dir, validate=False, config =['bids', 'derivatives'])
        
        if args.participant_label:
            input_subjects = args.participant_label
            subjects_to_analyze = layout.get_subjects(subject=input_subjects)
        else:
            subjects_to_analyze = layout.get_subjects()

        if not subjects_to_analyze:
            parser.error('Could not find participants. Please make sure the BIDS data '
                         'structure is present and correct. Datasets can be validated online '
                         'using the BIDS Validator (http://incf.github.io/bids-validator/).')

        if not args.atlas.endswith(('.nii', '.nii.gz')):
            parser.error('--atlas should end with .nii or .nii.gz')
        
        if args.bids_filter_file is not None:
            filter_list = json.loads(Path(args.bids_filter_file).read_text()) 
            
            default_input = {'extension': 'nii.gz', 
                             'datatype' : 'func', 
                             'desc': 'preproc', 
                             'task' : 'rest',
                             'suffix': 'bold'}
            default_input['subject']=subjects_to_analyze            
            default_input.update(filter_list['bold'])
            
            all_inputs = layout.get(return_type='filename',**default_input)
        
        else :
            all_inputs = layout.get(return_type='filename',datatype='func', subject=subjects_to_analyze, task='rest',desc='preproc',suffix='bold', extension=['nii', 'nii.gz'])
            
        if not all_inputs != []:
            parser.error('There are no files of type *bold.nii / *bold.nii.gz '
                         'Please make sure to have at least one file of the above type '
                         'in the BIDS specification')
        else:
            num_errors = 0
            for file_count in range(len(all_inputs)):
                try:
                    TR = layout.get_metadata(all_inputs[file_count])['RepetitionTime']
                except KeyError as e:
                    TR = spm_dep.spm.spm_vol(all_inputs[file_count]).header.get_zooms()[-1]
                para['TR'] = TR
                para['dt'] = para['TR'] / para['T']
                para['lag'] = np.arange(np.fix(para['min_onset_search'] / para['dt']),
                                        np.fix(para['max_onset_search'] / para['dt']) + 1,
                                        dtype='int')
                num_errors += 1
                try:
                    fourD_rsHRF.demo_rsHRF(all_inputs[file_count], args.atlas, args.output_dir, para, args.n_jobs, file_type, mode='bids w/ atlas', temporal_mask=temporal_mask, wiener=args.wiener)
                    num_errors -=1
                except ValueError as err:
                    print(err.args[0])
                except:
                    print("Unexpected error:", sys.exc_info()[0])
            success = len(all_inputs) - num_errors
            if success == 0:
                raise RuntimeError('Dimensions were inconsistent for all input-mask pairs; \n'
                                   'No inputs were processed!')

    if args.bids_dir is not None and args.brainmask:
        # carry analysis with bids_dir and brainmask
        layout = BIDSLayout(args.bids_dir, validate=False, config =['bids', 'derivatives'])

        if args.participant_label:
            input_subjects = args.participant_label
            subjects_to_analyze = layout.get_subjects(subject=input_subjects)
        else:
            subjects_to_analyze = layout.get_subjects()

        if not subjects_to_analyze:
            parser.error('Could not find participants. Please make sure the BIDS data '
                         'structure is present and correct. Datasets can be validated online '
                         'using the BIDS Validator (http://incf.github.io/bids-validator/).')

        if args.bids_filter_file is not None:
            filter_list = json.loads(Path(args.bids_filter_file).read_text()) 
            
            default_input = {'extension': 'nii.gz', 
                             'datatype' : 'func', 
                             'desc': 'preproc', 
                             'task' : 'rest',
                             'suffix': 'bold'}
            default_input['subject']=subjects_to_analyze            
            default_input.update(filter_list['bold'])
            
            all_inputs = layout.get(return_type='filename',**default_input)
            
            default_mask={'extension': 'nii.gz',
                          'datatype': 'func',
                          'desc': 'brain',
                          'task':'rest',
                          'suffix':'mask'}
            default_mask['subject']=subjects_to_analyze
            default_mask.update(filter_list['mask'])
            
            all_masks = layout.get(return_type='filename',**default_mask)
            
            
        else:            
            all_inputs = layout.get(return_type='filename',datatype='func', subject=subjects_to_analyze, task='rest',desc='preproc',suffix='bold', extension=['nii', 'nii.gz'])
            all_masks = layout.get(return_type='filename', datatype='func', subject=subjects_to_analyze, task='rest',desc='brain',suffix='mask', extension=['nii', 'nii.gz'])
       
        if not all_inputs != []:
            parser.error('There are no files of type *bold.nii / *bold.nii.gz '
                         'Please make sure to have at least one file of the above type '
                         'in the BIDS specification')
        if not all_masks != []:
            parser.error('There are no files of type *mask.nii / *mask.nii.gz '
                         'Please make sure to have at least one file of the above type '
                         'in the BIDS specification')
        if len(all_inputs) != len(all_masks):
            parser.error('The number of *bold.nii / .nii.gz and the number of '
                         '*mask.nii / .nii.gz are different. Please make sure that '
                         'there is one mask for each input_file present')

        all_inputs.sort()
        all_masks.sort()

        all_prefix_match = False
        prefix_match_count = 0
        for i in range(len(all_inputs)):
            input_prefix = all_inputs[i].split('/')[-1].split('_desc')[0]
            mask_prefix = all_masks[i].split('/')[-1].split('_desc')[0]
            if input_prefix == mask_prefix:
                prefix_match_count += 1
            else:
                all_prefix_match = False
                break
        if prefix_match_count == len(all_inputs):
            all_prefix_match = True

        if not all_prefix_match:
            parser.error('The mask and input files should have the same prefix for correspondence. '
                         'Please consider renaming your files')
        else:
            num_errors = 0
            for file_count in range(len(all_inputs)):
                file_type = all_inputs[file_count].split('bold')[1]
                if file_type == ".nii" or file_type == ".nii.gz":
                    try:
                        TR = layout.get_metadata(all_inputs[file_count])['RepetitionTime']
                    except KeyError as e:
                        TR = spm_dep.spm.spm_vol(all_inputs[file_count]).header.get_zooms()[-1]
                    para['TR'] = TR
                else:
                    spm_dep.spm.spm_vol(all_inputs[file_count])
                    TR = spm_dep.spm.spm_vol(all_inputs[file_count]).get_arrays_from_intent("NIFTI_INTENT_TIME_SERIES")[0].meta.get_metadata()["TimeStep"]
                    para['TR'] = float(TR) * 0.001


                para['dt'] = para['TR'] / para['T']
                para['lag'] = np.arange(np.fix(para['min_onset_search'] / para['dt']),
                                        np.fix(para['max_onset_search'] / para['dt']) + 1,
                                        dtype='int')
                num_errors += 1
                try:
                    fourD_rsHRF.demo_rsHRF(all_inputs[file_count], all_masks[file_count], args.output_dir, para, args.n_jobs, mode='bids', temporal_mask=temporal_mask, wiener=args.wiener)
                    num_errors -=1
                except ValueError as err:
                    print(err.args[0])
                except:
                    print("Unexpected error:", sys.exc_info()[0])
            success = len(all_inputs) - num_errors
            if success == 0:
                raise RuntimeError('Dimensions were inconsistent for all input-mask pairs; \n'
                                   'No inputs were processed!')



def main():
    warnings.filterwarnings("ignore")
    run_rsHRF()


if __name__ == '__main__':
    raise RuntimeError("CLI.py should not be run directly;\n"
                       "Please `pip install` rsHRF and use the `rsHRF` command")


import os
import matplotlib
matplotlib.use('agg')
import numpy             as np
import nibabel           as nib
import scipy.io          as sio
import matplotlib.pyplot as plt
from bids.layout  import BIDSLayout, parse_file_entities 
from scipy        import stats, signal
from scipy.sparse import lil_matrix
from rsHRF        import spm_dep, processing, parameters, basis_functions, utils, iterative_wiener_deconv

import warnings
warnings.filterwarnings("ignore")

def demo_rsHRF(input_file, mask_file, output_dir, para, p_jobs, file_type=".nii", mode="bids", wiener=False, temporal_mask=[]):
    # book-keeping w.r.t parameter values
    if 'localK' not in para or para['localK'] == None:
        if para['TR']<=2:
            para['localK'] = 1
        else:
            para['localK'] = 2
    # creating the output-directory if not already present
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
    # for four-dimensional input
    if mode != 'time-series':
        if mode == 'bids' or mode == 'bids w/ atlas':
            name = input_file.split('/')[-1].split('.')[0]
            v1 = spm_dep.spm.spm_vol(input_file)
        else:
            name = input_file.split('/')[-1].split('.')[0]   
            v1 = spm_dep.spm.spm_vol(input_file)
        if mask_file != None:
            if mode == 'bids':
                mask_name = mask_file.split('/')[-1].split('.')[0]
                v = spm_dep.spm.spm_vol(mask_file)
            else:
                mask_name = mask_file.split('/')[-1].split('.')[0]
                v = spm_dep.spm.spm_vol(mask_file)
            if file_type == ".nii" or file_type == ".nii.gz":
                brain = spm_dep.spm.spm_read_vols(v)
            else:
                brain = v.agg_data().flatten(order='F')
            if  ((file_type == ".nii" or file_type == ".nii.gz") and \
                    v1.header.get_data_shape()[:-1] != v.header.get_data_shape()) or \
                ((file_type == ".gii" or file_type == ".gii.gz") and \
                    v1.agg_data().shape[0]!= v.agg_data().shape[0]):
                raise ValueError ('Inconsistency in input-mask dimensions' + '\n\tinput_file == ' + name + file_type + '\n\tmask_file == ' + mask_name + file_type)
            else:
                if file_type == ".nii" or file_type == ".nii.gz" :
                    data   = v1.get_data()      
                else:
                    data   = v1.agg_data()
        else:
            print('No atlas provided! Generating mask file...')
            if file_type == ".nii" or file_type == ".nii.gz" :
                data   = v1.get_data() 
                brain  = np.nanvar(data.reshape(-1, data.shape[3]), -1, ddof=0)
            else:
                data   = v1.agg_data()
                brain  = np.nanvar(data, -1, ddof=0) 
            print('Done')
        voxel_ind  = np.where(brain > 0)[0]
        mask_shape = data.shape[:-1]
        nobs       = data.shape[-1]
        data1      = np.reshape(data, (-1, nobs), order='F').T
        bold_sig = stats.zscore(data1[:, voxel_ind], ddof=1)
   # for time-series input
    else:
        name = input_file.split('/')[-1].split('.')[0]
        data1 = (np.loadtxt(input_file, delimiter=","))
        if data1.ndim == 1:
            data1 = np.expand_dims(data1, axis=1)
        nobs = data1.shape[0]
        bold_sig = stats.zscore(data1, ddof=1)
    if len(temporal_mask) > 0 and len(temporal_mask) != nobs:
            raise ValueError ('Inconsistency in temporal_mask dimensions.\n' + 'Size of mask: ' + str(len(temporal_mask)) + '\n' + 'Size of time-series: ' + str(nobs))
    bold_sig = np.nan_to_num(bold_sig)
    bold_sig_deconv = processing. \
                      rest_filter. \
                      rest_IdealFilter(bold_sig, para['TR'], para['passband_deconvolve'])   
    bold_sig = processing. \
               rest_filter. \
               rest_IdealFilter(bold_sig, para['TR'], para['passband'])   
    data_deconv  = np.zeros(bold_sig.shape)
    event_number = np.zeros((1, bold_sig.shape[1]))
    print('Retrieving HRF ...')
    #Estimate HRF for the fourier / hanning / gamma / cannon basis functions
    if not (para['estimation'] == 'sFIR' or para['estimation'] == 'FIR'):
        bf = basis_functions.basis_functions.get_basis_function(bold_sig.shape, para)
        beta_hrf, event_bold = utils.hrf_estimation.compute_hrf(bold_sig, para, temporal_mask, p_jobs, bf=bf)
        hrfa = np.dot(bf, beta_hrf[np.arange(0, bf.shape[1]), :])
    #Estimate HRF for FIR and sFIR
    else:
        para['T'] = 1        
        beta_hrf, event_bold = utils.hrf_estimation.compute_hrf(bold_sig, para, temporal_mask, p_jobs)
        hrfa = beta_hrf[:-1,:]
    nvar = hrfa.shape[1]
    PARA = np.zeros((3, nvar))
    for voxel_id in range(nvar):
        hrf1 = hrfa[:, voxel_id]
        PARA[:, voxel_id] = \
            parameters.wgr_get_parameters(hrf1, para['TR'] / para['T'])
    print('Done')
    print('Deconvolving HRF ...')
    if para['T'] > 1:
        hrfa_TR = signal.resample_poly(hrfa, 1, para['T'])
    else:
        hrfa_TR = hrfa
    for voxel_id in range(nvar):
        hrf = hrfa_TR[:, voxel_id]
        if not wiener:
            H = np.fft.fft(
                np.append(hrf,
                          np.zeros((nobs - max(hrf.shape), 1))), axis=0)
            M = np.fft.fft(bold_sig_deconv[:, voxel_id])
            data_deconv[:, voxel_id] = \
                np.fft.ifft(H.conj() * M / (H * H.conj() + .1*np.mean((H * H.conj()))))
        else:
            data_deconv[:, voxel_id] = iterative_wiener_deconv.rsHRF_iterative_wiener_deconv(bold_sig_deconv[:, voxel_id], hrf)
        event_number[:, voxel_id] = np.amax(event_bold[voxel_id].shape)
    print('Done')
    print('Saving Output ...')
    # setting the output-path
    if mode == 'bids' or mode == 'bids w/ atlas':
        layout_output = BIDSLayout(output_dir)
        entities = parse_file_entities(input_file)
        sub_save_dir = layout_output.build_path(entities).rsplit('/',1)[0]            
    else:
        sub_save_dir = output_dir
    if not os.path.isdir(sub_save_dir):
        os.makedirs(sub_save_dir, exist_ok=True)
    dic = {'para': para, 'hrfa': hrfa, 'event_bold': event_bold, 'PARA': PARA}
    ext = '_hrf.mat'
    if mode == "time-series":
        dic["event_number"] = event_number
        dic["data_deconv"]  = data_deconv
        ext = '_hrf_deconv.mat'
    name = name.rsplit('_bold', 1)[0]   
    sio.savemat(os.path.join(sub_save_dir, name + ext), dic)
    HRF_para_str = ['height', 'T2P', 'FWHM']
    if mode != "time-series":
        mask_data = np.zeros(mask_shape).flatten(order='F')
        for i in range(3):
            fname = os.path.join(sub_save_dir,
                                 name + '_' + HRF_para_str[i])
            mask_data[voxel_ind] = PARA[i, :]
            mask_data = mask_data.reshape(mask_shape, order='F')
            spm_dep.spm.spm_write_vol(v1, mask_data, fname, file_type)
            mask_data = mask_data.flatten(order='F')
        fname = os.path.join(sub_save_dir, name + '_eventnumber')
        mask_data[voxel_ind] = event_number
        mask_data = mask_data.reshape(mask_shape, order='F')
        spm_dep.spm.spm_write_vol(v1, mask_data, fname, file_type)
        mask_data = np.zeros(data.shape)
        dat3 = np.zeros(data.shape[:-1]).flatten(order='F')
        for i in range(nobs):
            fname = os.path.join(sub_save_dir, name + '_deconv')
            dat3[voxel_ind] = data_deconv[i, :]
            dat3 = dat3.reshape(data.shape[:-1], order='F')
            if file_type == ".nii" or file_type == ".nii.gz" :
                mask_data[:, :, :, i] = dat3
            else:
                mask_data[:, i] = dat3
            dat3 = dat3.flatten(order='F')
        spm_dep.spm.spm_write_vol(v1, mask_data, fname, file_type)
    pos = 0
    while pos < hrfa_TR.shape[1]:
        if np.any(hrfa_TR[:,pos]):
            break 
        pos += 1
    event_plot = lil_matrix((1, nobs))
    if event_bold.size:
        event_plot[:, event_bold[pos]] = 1
    else:
        print("No Events Detected!")
        return 0
    event_plot = np.ravel(event_plot.toarray())
    plt.figure()
    plt.plot(para['TR'] * np.arange(1, np.amax(hrfa_TR[:, pos].shape) + 1),
             hrfa_TR[:, pos], linewidth=1)
    plt.xlabel('time (s)')
    plt.savefig(os.path.join(sub_save_dir, name + '_hrf_plot.png'))
    plt.figure()
    plt.plot(para['TR'] * np.arange(1, nobs + 1),
             np.nan_to_num(stats.zscore(bold_sig[:, pos], ddof=1)),
             linewidth=1)
    plt.plot(para['TR'] * np.arange(1, nobs + 1),
             np.nan_to_num(stats.zscore(data_deconv[:, pos], ddof=1)),
             color='r', linewidth=1)
    markerline, stemlines, baseline = \
        plt.stem(para['TR'] * np.arange(1, nobs + 1), event_plot)
    plt.setp(baseline, 'color', 'k', 'markersize', 1)
    plt.setp(stemlines, 'color', 'k')
    plt.setp(markerline, 'color', 'k', 'markersize', 3, 'marker', 'd')
    plt.legend(['BOLD', 'Deconvolved BOLD', 'Events'], loc='best')
    plt.xlabel('time (s)')
    plt.savefig(os.path.join(sub_save_dir, name + '_deconvolution_plot.png'))   
    print('Done')
    return 0



import numpy as np
from scipy.io   import savemat
from copy       import deepcopy

from ...datatypes.misc.parameters import Parameters 
from .timeseries                  import TimeSeries
from .bold_raw                    import BOLD_Raw

class BOLD_Preprocessed(TimeSeries):
    """
    This stores the Preprocessed BOLD Time-series
    
    Attributes:
        1. BOLD_Raw     : The Raw BOLD time-series object through which it was derived
        2. mask_file    : The mask-file path
    """
    def __init__(self, label="",ts=np.array([]),subject_index="", para=Parameters()):
        TimeSeries.__init__(self, label="",ts=np.array([]),subject_index="", para=Parameters())
        self.label          = label 
        self.subject_index  = subject_index 
        self.timeseries     = ts 
        self.shape          = ts.shape
        self.parameters     = deepcopy(para) 
        self.BOLD_Raw       = BOLD_Raw()
        self.mask_file      = ""

    # setters
    def set_maskfile(self, mask_file):
        self.mask_file = mask_file 

    def set_BOLD_Raw(self, BOLD_Raw):
        self.BOLD_Raw = BOLD_Raw
    
    # getters
    def get_maskfile(self):
        return self.mask_file 

    def get_BOLD_Raw(self):
        return self.BOLD_Raw

    # misc.
    def compareTimeSeries(self, ts):
        """ Compares another time-series with itself to determine if both are identical 
            Four checks are performed:
                1. Label
                2. Parameters
                3. Raw BOLD associated with it
                4. Mask-file
            If all the three comparisions return true, then both the HRF
                time-series objects are identical
        """
        if self.label == ts.get_label() \
            and self.parameters == ts.get_parameters() \
            and self.BOLD_Raw.compareTimeSeries(ts.get_BOLD_Raw()) \
            and ts.get_maskfile() == self.mask_file:
            return True 
        else:
            return False 
            
    def save_info(self, name):
        """ Saves the information about the time-series in a .mat file """
        try:
            dic = {}
            dic["timeseries"] = self.timeseries
            dic["mask_file"] = self.mask_file 
            savemat(name, dic)
            return True
        except:
            return False


import re
import os
import sys
from setuptools import setup, find_packages
from setuptools.command.install import install

with open("rsHRF/VERSION", "r") as fh:
    VERSION = fh.read().strip('\n')

with open("README.md", "r") as fh:
    long_description = fh.read()

class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)    
    
setup(
    name="rsHRF",
    packages=find_packages(),
    entry_points={
        "console_scripts": ['rsHRF = rsHRF.CLI:main']
    },
    version=VERSION,
    description="BIDs App to retrieve the haemodynamic response function from resting state fMRI data",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Madhur Tandon, Amogh Johri",
    author_email="madhurtandon23@gmail.com, amogh.johri@iiitb.org",
    url="https://github.com/BIDS-Apps/rsHRF",
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    install_requires=["numpy", "nibabel", "matplotlib", "scipy", "pybids==0.11.1", "pandas", "patsy", "mpld3", "duecredit", "joblib", "pyyawt"],
    cmdclass={
        'verify': VerifyVersionCommand,
    },
)


import numpy as np
import warnings

warnings.filterwarnings("ignore")

def knee_pt(y):
    res_x, _id = knee_pt_helper(y)
    _idm = np.argmin(y)
    ratio = np.abs(y[_id]-y[_idm])/np.abs(np.max(y) - np.min(y))
    if ratio > 0.5:
        idx_of_result = _idm 
    else:
        idx_of_result = _id 
    return res_x, idx_of_result

def knee_pt_helper(y, x=None):
    x_was_none = False
    use_absolute_dev_p = True
    res_x = np.nan
    idx_of_result = np.nan

    if type(y) is not np.ndarray:
        print('knee_pt: y must be a numpy 1D vector')
        return res_x, idx_of_result
    else:
        if y.ndim >= 2:
            print('knee_pt: y must be 1 dimensional')
            return res_x, idx_of_result
        if np.size(y) == 0:
            print('knee_pt: y can not be an empty vector')
            return res_x, idx_of_result
        else:
            if x is None:
                x_was_none = True
                x = np.arange(1, np.amax(y.shape) + 1, dtype=np.int)
            if x.shape != y.shape:
                print('knee_pt: y and x must have the same dimensions')
                return res_x, idx_of_result
            if y.size < 3:
                res_x, idx_of_result = np.min(y), np.argmin(y)
                return res_x, idx_of_result
            if np.all(np.diff(x) >= 0) and (not x_was_none):
                idx = np.argsort(x)
                y = np.sort(y)
                x = np.sort(x)
            else:
                idx = np.arange(0, np.amax(x.shape))
            sigma_xy = np.cumsum(np.multiply(x, y), axis=0)
            sigma_x = np.cumsum(x, axis=0)
            sigma_y = np.cumsum(y, axis=0)
            sigma_xx = np.cumsum(np.multiply(x, x), axis=0)
            n = np.arange(1, np.amax(y.shape) + 1).conj().T
            det = np.multiply(n, sigma_xx) - np.multiply(sigma_x, sigma_x)
            mfwd = (np.multiply(n, sigma_xy) -
                    np.multiply(sigma_x, sigma_y)) / det
            bfwd = -1 * ((np.multiply(sigma_x, sigma_xy) -
                          np.multiply(sigma_xx, sigma_y)) / det)

            sigma_xy = np.cumsum(np.multiply(x[::-1], y[::-1]), axis=0)
            sigma_x = np.cumsum(x[::-1], axis=0)
            sigma_y = np.cumsum(y[::-1], axis=0)
            sigma_xx = np.cumsum(np.multiply(x[::-1], x[::-1]), axis=0)
            n = np.arange(1, np.amax(y.shape) + 1).conj().T
            det = np.multiply(n, sigma_xx) - np.multiply(sigma_x, sigma_x)
            mbck = ((np.multiply(n, sigma_xy) -
                     np.multiply(sigma_x, sigma_y)) / det)[::-1]
            bbck = (-1 *
                    ((np.multiply(sigma_x, sigma_xy) -
                      np.multiply(sigma_xx, sigma_y)) / det))[::-1]

            error_curve = np.full(y.shape, np.nan)
            for breakpt in range(1, np.amax((y - 1).shape)):
                delsfwd = (np.multiply(mfwd[breakpt], x[0:breakpt + 1]) +
                           bfwd[breakpt]) - y[0:breakpt + 1]
                delsbck = (np.multiply(mbck[breakpt], x[breakpt:]) +
                           bbck[breakpt]) - y[breakpt:]
                if use_absolute_dev_p:
                    error_curve[breakpt] = \
                        np.sum(np.abs(delsfwd)) + np.sum(np.abs(delsbck))
                else:
                    error_curve[breakpt] = \
                        np.sqrt(np.sum(np.multiply(delsfwd, delsfwd))) + \
                        np.sqrt(np.sum(np.multiply(delsbck, delsbck)))
            try:
                loc = np.nanargmin(error_curve)
            except ValueError as e:
                loc = 0
            res_x = x[loc]
            idx_of_result = idx[loc]
    return res_x, idx_of_result
